
#include "CapriccioAI.h"

#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <stdio.h>

#include "BoardStonePointEnumerator.h"

static const char JosekiFileName[] = "9x9AIJoseki.sgf";

// 後でファイル化する
static const unsigned int patterns[] = 
{
	0x00010400,
	// + + +
	// + O *
	// * X +
	0x00000401,
	// * X +
	// + O *
	// + + +
	0x00040100,
	// + + +
	// * O +
	// + X *
	0x00000104,
	// + X *
	// * O +
	// + + +
	0x00040002,
	// + * +
	// + O X
	// + + *
	0x00010002,
	// + * +
	// X O +
	// * + +
	0x00020004,
	// + + *
	// + O X
	// + * +
	0x00020001,
	// * + +
	// X O +
	// + * +
	0x00020002,
	// + * +
	// + O X
	// + * +
	0x00020002,
	// + * +
	// X O +
	// + * +
	0x00000500,
	// + + +
	// * O *
	// + X +
	0x00000500,
	// + X +
	// * O *
	// + + +

};

static const int numPatterns = sizeof(patterns) / sizeof (unsigned int);

static const char patternMatchCandidateOffsetX[] =
{
	0,
	0,
	0,
	0,
	1,
	-1,
	1,
	-1,
	1,
	-1,
	0,
	0,
};

static const char patternMatchCandidateOffsetY[] =
{
	1,
	-1,
	1,
	-1,
	0,
	0,
	0,
	0,
	0,
	0,
	1,
	-1,
};

CapriccioAI::CapriccioAI()
{
	srand((unsigned int)time(NULL));
	joseki_ = new JosekiDictionary(JosekiFileName);

	prevMove_.point.x = 0;
	prevMove_.point.y = 0;

	for (int i = 0; i < MAX_UNDOS; i++)
	{
		chainState_[i] = NONE;
	}
	
	searchList_.count_ = 0;

	// fill distance mask
	for (int x = 1; x <= 9; x++)
	{
		for (int y = 1; y <= 9; y++)
		{
			for (int dist = 0; dist < MAX_RELATIVE_DISTANCE; dist++)
			{
				int tempDist = dist + 1;
				int maskX, maskY;

				for (int i = 0; i < tempDist; i++)
				{
					maskX = x - tempDist + i;
					maskY = y - i;
					if (maskX >= 1 && maskX <= 9 && maskY >= 1 && maskY <= 9)
						distanceMask_[x][y][dist].SetStone(Point(maskX, maskY));

					maskX = x + tempDist - i;
					maskY = y + i;
					if (maskX >= 1 && maskX <= 9 && maskY >= 1 && maskY <= 9)
						distanceMask_[x][y][dist].SetStone(Point(maskX, maskY));
					
					maskX = x - i;
					maskY = y + tempDist - i;
					if (maskX >= 1 && maskX <= 9 && maskY >= 1 && maskY <= 9)
						distanceMask_[x][y][dist].SetStone(Point(maskX, maskY));
					
					maskX = x + i;
					maskY = y - tempDist + i;
					if (maskX >= 1 && maskX <= 9 && maskY >= 1 && maskY <= 9)
						distanceMask_[x][y][dist].SetStone(Point(maskX, maskY));
				}
			}
		}
	}

	// init 3-3 point move masks
	upperLeft33Mask_.Clear();
	upperRight33Mask_.Clear();
	lowerLeft33Mask_.Clear();
	lowerRight33Mask_.Clear();

	for (int x = 1; x <= 4; x++)
	{
		for (int y = 1; y <= 4; y++)
		{
			if (x == 4 && y == 4)
				break;

			upperLeft33Mask_. SetStone(Point(     x,     y));
			upperRight33Mask_.SetStone(Point(10 - x,     y));
			lowerLeft33Mask_. SetStone(Point(     x,10 - y));
			lowerRight33Mask_.SetStone(Point(10 - x,10 - y));
		}
	}
}

CapriccioAI::~CapriccioAI()
{
	delete joseki_;
}

void CapriccioAI::ClearBoard()
{
	manager_.Clear();

	prevMove_.point.x = 0;
	prevMove_.point.y = 0;

	for (int i = 0; i < MAX_UNDOS; i++)
	{
		chainState_[i] = NONE;
	}

	joseki_->Initialize();
}

void CapriccioAI::SetMove(Move move)
{
	manager_.SetStone(move);
	prevMove_ = move;

	joseki_->Proceed(manager_.GetCurrentBoard());
}

inline void SetMoveString(Move move, char* moveStr)
{
	moveStr[0] = 'a' + move.point.x + (move.point.x == 9 ? 1 : 0) - 1;
	moveStr[1] = '1' + move.point.y - 1;
	moveStr[2] = 0;
}

void CapriccioAI::NextMove(int color, char* moveStr)
{
	//clock_t startClock;
	//clock_t stopClock;

	//startClock = clock();

	Move move;
	move.color = color;

	move.point = joseki_->GetNextMove();
	if (move.point.x != 0)	 // if found
	{
		manager_.SetStone(move);
		SetMoveString(move, moveStr);
		prevMove_ = move;

		joseki_->Proceed(manager_.GetCurrentBoard());

		return;
	}

	// pattern match for immediate moves
	if (GetImmediateMove(move))
	{
		manager_.SetStone(move);
		SetMoveString(move, moveStr);
		prevMove_ = move;

		return;
	}

	//stopClock = clock();
	//printf("Joseki & Immediate: %.4lf sec\n", (stopClock - startClock)/(double)CLOCKS_PER_SEC);

	searchList_.count_ = 0;
	UpdateChainState();

	possibleMoveList_.count_ = 0;
	AppendPossibleMoves(possibleMoveList_, color);

	// clear relativity array
	for (int x = 1; x <= 9; x++)
	{
		for (int y = 1; y <= 9; y++)
		{
			relativity_[x][y] = 0;
		}
	}

	// init moveValue data
	int lambda = -1;
	int maxRelativity = -1;
	memset(moveValues_, 0 , BOARD_HEIGHT * BOARD_WIDTH * sizeof(int));
	// temp Variable for going through the list
	Move tempMove;
	int  tempLambda;
	int  tempRelativity;
	StoneChain *tempChain;

	// calculate relativity
	Point relativePt;
	StoneChainEnumerator enumerator;
	StoneChain* chain;
	for (int i = 0; i < possibleMoveList_.count_; i++)
	{
		relativePt = possibleMoveList_.moves_[i].move_.point;

		enumerator = manager_.GetEnumerator();
		while (chain = enumerator.Next())
		{
			if (chain->GetColor() == color)
			{
				relativity_[relativePt.x][relativePt.y] += GetRelativity(relativePt, chain);
			}
		}
	}

	// look for sake making/breaking moves
	searchList_.count_ = 0;
	FindSafeMakingMoves(searchList_, color);
	for (int i = 0; i < searchList_.count_; i++)
	{
		tempMove = searchList_.moves_[i].move_;
		tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += searchList_.moves_[i].value_);
		tempRelativity = relativity_[tempMove.point.x][tempMove.point.y];

		if (!manager_.IsSafe(tempMove))
			continue;

		if ( tempLambda > lambda ||
			(tempLambda == lambda && tempRelativity > maxRelativity))
		{
			move   = tempMove;
			lambda = tempLambda;
			maxRelativity = tempRelativity;
		}
	}
	searchList_.count_ = 0;
	FindSafeBreakingMoves(searchList_, color);
	for (int i = 0; i < searchList_.count_; i++)
	{
		tempMove = searchList_.moves_[i].move_;
		tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += searchList_.moves_[i].value_);
		tempRelativity = relativity_[tempMove.point.x][tempMove.point.y];

		tempLambda = (tempLambda >> 1) + (tempRelativity >> 1);

		if ( tempLambda > lambda ||
			(tempLambda == lambda && tempRelativity > maxRelativity))
		{
			move   = tempMove;
			lambda = tempLambda;
			maxRelativity = tempRelativity;
		}
	}

	//stopClock = clock();
	//printf("Safe Making: %.4lf sec\n", (stopClock - startClock)/(double)CLOCKS_PER_SEC);

	BitBoard9 libs;
	int numLibs;
	NeighborStoneChainEnumerator neighborEnumerator;
	StoneChain* neighbor;
	BitBoard9 neighborLibs, newLibs;
	int numNewLibs;
	BoardStonePointEnumerator libEnumerator;
	BitBoard9 tempBoard1, tempBoard2;
	Move libMove, oppMove;
	bool continueLoop;
	bool hasWeakNeighbor;
	int  tempListCount;
	int  count;

	enumerator = manager_.GetEnumerator();

	while ( chain = enumerator.Next() )
	{
		if (chainState_[chain->GetIndex()] & ( WEAK | SAFE ))
		{
			continue;
		}
		else if ( chain->GetColor() == move.color )
		{

			searchList_.count_ = 0;
			manager_.GetLiberties(chain, &libs);
			numLibs = libs.CountStones();

			if (numLibs == 1)
			{
				FindDefenseMove(searchList_, chain, 1);
			}
			else
			{
				if (((chainState_[chain->GetIndex()] & URGENT) != 0) && IsLooseLadder(chain))
				{
					continue;
				}
				else if (numLibs == 2)
				{
					// if both libs connects to the same chain, no need to defend
					libEnumerator.Initialize(libs);
					libMove.color = move.color;

					libMove.point = *(libEnumerator.Next());

					if (manager_.IsLegalForBothColor(libMove.point))
					{
						manager_.SetStone(libMove);
						tempBoard1 = manager_.GetNewestStoneChain()->GetStones();
						manager_.Undo();
						
						libMove.point = *(libEnumerator.Next());

						if (manager_.IsLegalForBothColor(libMove.point))
						{
							manager_.SetStone(libMove);
							tempBoard2 = manager_.GetNewestStoneChain()->GetStones();
							manager_.Undo();

							if (tempBoard2.CountStones() != chain->GetNumStones() + 1)
							{
								tempBoard2 |= tempBoard1;
								tempBoard2 -= tempBoard1;

								if (tempBoard2.CountStones() == 1)
								{
									continue;
								}
							}
						}
					}
				}

				hasWeakNeighbor = false;

				neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);
				while (neighbor = neighborEnumerator.Next())
				{
					//if (chainState_[neighbor->GetIndex()] & (WEAK | URGENT))
					if (chainState_[neighbor->GetIndex()] & WEAK)
					{
						hasWeakNeighbor = true;
						break;
					}
				}

				// perform shallower search if there exists a weak neighbor
				if (hasWeakNeighbor)
					FindDefenseMove(searchList_, chain, 3);
				else 
					FindDefenseMove(searchList_, chain, 4);
			}

			for (int i = 0; i < searchList_.count_; i++)
			{
				tempMove = searchList_.moves_[i].move_;
				tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += searchList_.moves_[i].value_);
				tempRelativity = relativity_[tempMove.point.x][tempMove.point.y];

				// try the move to see if the move hasn't made the chain urgent
				manager_.SetStone(tempMove);
				tempChain = manager_.GetNewestStoneChain();
				if (!tempChain->GetStones().HasStonesInCommon(chain->GetStones()))
				{
					tempChain = chain;
				}
				if (CheckIfUrgent(tempChain))
				{
					// consider making own chain a semi-suicide
					manager_.Undo();
					continue;
				}
				else if (chainState_[chain->GetIndex()] & URGENT)
				{
					// released from the urgent state
					tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] +=chain->GetNumStones() << 1);
				}
				//else
				//{
				//	// value def moves bit high
				//	tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] +=chain->GetNumStones());
				//}

				manager_.Undo();

				// value less the move such that the opponent cannot immediately be played.
				oppMove = tempMove;
				oppMove.color = tempMove.color ^ 0x01;

				if (!IsStrictlySafe(oppMove))
				{
					tempLambda = tempLambda >> 1;
				}

				if ( tempLambda > lambda ||
					(tempLambda == lambda && tempRelativity > maxRelativity))
				{
					move   = tempMove;
					lambda = tempLambda;
					maxRelativity = tempRelativity;
				}
			}
		}
		else
		{
			// 大石死なず
			manager_.GetLiberties(chain, &libs);
			numLibs = libs.CountStones();
			if (chain->GetNumStones() > 4 && numLibs > 3)
				continue;

			if (numLibs > 1 && 0 == (chainState_[chain->GetIndex()] & URGENT))
			{
				// check if the stones involved in the attack is not obviously weak
				continueLoop = false;
				tempListCount = searchList_.count_;
				neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);

				if (numLibs == 2)
				{
					// if the prospective attack move is an atari
					while (neighbor = neighborEnumerator.Next())
					{
						manager_.GetLiberties(neighbor, &neighborLibs);

						if (0 != (chainState_[neighbor->GetIndex()] & WEAK) || neighborLibs.CountStones() <= 1 )
						{
							continueLoop = true;
							break;
						}
					}
				}
				else
				{
					// otherwise...
					while (neighbor = neighborEnumerator.Next())
					{
						if (0 != (chainState_[neighbor->GetIndex()] & (WEAK | URGENT)) )
						{
							continueLoop = true;
							break;
						}
					}
				}
				searchList_.count_ = tempListCount;

				if (continueLoop)
					continue;
			}
			else if (0 != (chainState_[chain->GetIndex()] & URGENT) && IsLooseLadder(chain))
			{
				continueLoop = true;
				neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);

				while (neighbor = neighborEnumerator.Next())
				{
					if (chainState_[neighbor->GetIndex()] & URGENT)
					{
						continueLoop = false;
						break;
					}
				}
				
				if (continueLoop)
					continue;
			}

			searchList_.count_ = 0;

			FindAttackMove(searchList_, chain, 6) ;

			for (int i = 0; i < searchList_.count_; i++)
			{
				tempMove = searchList_.moves_[i].move_;
				tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += searchList_.moves_[i].value_);
				tempRelativity = relativity_[tempMove.point.x][tempMove.point.y];

				if (numLibs > 2 && !IsStrictlySafe(tempMove))
				{
					continue;
				}

				manager_.SetStone(tempMove);
				// making the opponent's chain urgent willl increase the value of the move

				// test ... avoiding hasami-tsuke
				manager_.GetLiberties(manager_.GetNewestStoneChain(), &newLibs);
				numNewLibs = newLibs.CountStones();

				if (numLibs > 2 && tempRelativity <= 0 && numNewLibs < 4)
				{
					manager_.Undo();
					continue;
				}

				if ((0 == (chainState_[chain->GetIndex()] & URGENT)) && CheckIfUrgent(chain))
				{
					if (IsLooseLadder(chain))
						tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += searchList_.moves_[i].value_ + chain->GetNumStones());
					else
						tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += searchList_.moves_[i].value_);
				}
				else if (CheckIfWeak(chain))
				{
					tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] = (tempLambda << 1));
				}

				// check for double atari
				if (numLibs == 2)
				{
					neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);
					count = 0;

					// if the prospective attack move is an atari
					while (neighbor = neighborEnumerator.Next())
					{
						manager_.GetLiberties(neighbor, &neighborLibs);

						if ( neighborLibs.CountStones() == 1 )
						{
							if (count > 0)
							{
								tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += (count + neighbor->GetNumStones()));
								break;
							}

							count += neighbor->GetNumStones();
						}
					}
				}

				manager_.Undo();

				if ( tempLambda > lambda ||
					(tempLambda == lambda && tempRelativity > maxRelativity))
				{
					move   = tempMove;
					lambda = tempLambda;
					maxRelativity = tempRelativity;
				}
			}
		}
	}

	//stopClock = clock();
	//printf("Lambda: %.4lf sec\n", (stopClock - startClock)/(double)CLOCKS_PER_SEC);

	// try corner moves if the attack & defense search did not find a move
	if (lambda < 0)
	{
		if (manager_.IsEmpty(upperLeft33Mask_))
		{
			move.point = Point(3,3);
			lambda = 0;
		}
		else if (manager_.IsEmpty(upperRight33Mask_))
		{
			move.point = Point(7,3);
			lambda = 0;
		}
		else if (manager_.IsEmpty(lowerLeft33Mask_))
		{
			move.point = Point(3,7);
			lambda = 0;
		}
		else if (manager_.IsEmpty(lowerRight33Mask_))
		{
			move.point = Point(7,7);
			lambda = 0;
		}
	}

	// yose move
	if (lambda < 0)
	{

		// other yose moves
		if (lambda < 0)
		{
			searchList_.count_ = 0;

			FindYoseMoves(searchList_, color);

			for (int i = 0; i < searchList_.count_; i++)
			{
				tempMove = searchList_.moves_[i].move_;
				tempLambda = (moveValues_[tempMove.point.x][tempMove.point.y] += searchList_.moves_[i].value_);
				tempRelativity = relativity_[tempMove.point.x][tempMove.point.y];

				if ( tempLambda > lambda ||
					(tempLambda == lambda && tempRelativity > maxRelativity))
				{
					move   = tempMove;
					lambda = tempLambda;
					maxRelativity = tempRelativity;
				}
			}
		}
	}


	if (lambda >= 0)
	{
		manager_.SetStone(move);
		SetMoveString(move, moveStr);
		prevMove_ = move;

		return;
	}


	// random move

	//if (possibleMoveList_.count_ > 0)
	//{
	//	move = possibleMoveList_.moves_[rand() % possibleMoveList_.count_].move_;

	//	manager_.SetStone(move);
	//	SetMoveString(move, moveStr);
	//	prevMove_ = move;

	//	return;
	//}
	
	move.point.x = 0;
	move.point.y = 0;
	manager_.SetStone(move);
	prevMove_ = move;
	strncpy_s( moveStr, 16, "pass", 4);
}

bool CapriccioAI::WasPrevMoveAtari()
{
	return manager_.HasAtariChainAround(prevMove_);
}

bool CapriccioAI::GetImmediateMove(Move& move)
{
	if (prevMove_.point.x > 0 && !WasPrevMoveAtari())
	{
		if (prevMove_.point.x > 1 && prevMove_.point.x < 9 && prevMove_.point.y > 1 &&prevMove_.point.y < 9)
		{
			// pattern match
			for (int i = 0; i < numPatterns; i++)
			{
				if (manager_.Match3x3Pattern(patterns[i], prevMove_))
				{
					move.point.x = prevMove_.point.x + patternMatchCandidateOffsetX[i];
					move.point.y = prevMove_.point.y + patternMatchCandidateOffsetY[i];

					if (manager_.IsSafe(move))
					{
						return true;
					}
				}
			}
		}

		// find response move to low hane moves
		return GetYoseHaneResponseMove(move);
	}

	return false;
}

bool CapriccioAI::GetYoseHaneResponseMove(Move& move)
{
	int prevX = prevMove_.point.x;
	int prevY = prevMove_.point.y;

	StoneChain* chain;

	// hane on the second line
	if (prevX == 2 && prevY > 2 && prevY < 8)
	{
		chain = manager_.GetChainAt(Point(3, prevY));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckSecondLineOsae(move, Move(move.color, 2, prevY + 1), Move(move.color ^ 0x01, 3, prevY + 1), Move(move.color ^ 0x01, 4, prevY), Point(3, prevY)) ||
			    CheckSecondLineOsae(move, Move(move.color, 2, prevY - 1), Move(move.color ^ 0x01, 3, prevY - 1), Move(move.color ^ 0x01, 4, prevY), Point(3, prevY)) );
	}
	else if (prevX == 8 && prevY > 2 && prevY < 8)
	{
		chain = manager_.GetChainAt(Point(7, prevY));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckSecondLineOsae(move, Move(move.color, 8, prevY + 1), Move(move.color ^ 0x01, 7, prevY + 1), Move(move.color ^ 0x01, 6, prevY), Point(7, prevY)) ||
			    CheckSecondLineOsae(move, Move(move.color, 8, prevY - 1), Move(move.color ^ 0x01, 7, prevY - 1), Move(move.color ^ 0x01, 6, prevY), Point(7, prevY)) );
	}
	else if (prevY == 2 && prevX > 2 && prevX < 8)
	{
		chain = manager_.GetChainAt(Point(prevX, 3));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckSecondLineOsae(move, Move(move.color, prevX + 1, 2), Move(move.color ^ 0x01, prevX + 1, 3), Move(move.color ^ 0x01, prevX, 4), Point(prevX, 3)) ||
			    CheckSecondLineOsae(move, Move(move.color, prevX - 1, 2), Move(move.color ^ 0x01, prevX - 1, 3), Move(move.color ^ 0x01, prevX, 4), Point(prevX, 3)) );
	}
	else if (prevY == 8 && prevX > 2 && prevX < 8)
	{
		chain = manager_.GetChainAt(Point(prevX, 7));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckSecondLineOsae(move, Move(move.color, prevX + 1, 8), Move(move.color ^ 0x01, prevX + 1, 7), Move(move.color ^ 0x01, prevX, 6), Point(prevX, 7)) ||
			    CheckSecondLineOsae(move, Move(move.color, prevX - 1, 8), Move(move.color ^ 0x01, prevX - 1, 7), Move(move.color ^ 0x01, prevX, 6), Point(prevX, 7)) );
	}
	// hane on the first line
	else if (prevX == 1 && prevY > 2 && prevY < 8)
	{
		chain = manager_.GetChainAt(Point(2, prevY));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckFirstLineYoseHane(move, Point(1, prevY + 1), Point(1, prevY - 1), Point(2, prevY + 1), Point(2, prevY - 1)));
	}
	else if (prevX == 9 && prevY > 2 && prevY < 8)
	{
		chain = manager_.GetChainAt(Point(8, prevY));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckFirstLineYoseHane(move, Point(9, prevY + 1), Point(9, prevY - 1), Point(8, prevY + 1), Point(8, prevY - 1)));
	}
	else if (prevY == 1 && prevX > 2 && prevX < 8)
	{
		chain = manager_.GetChainAt(Point(prevX, 2));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckFirstLineYoseHane(move, Point(prevX + 1, 1), Point(prevX - 1, 1), Point(prevX + 1, 2), Point(prevX - 1, 2)));
	}
	else if (prevY == 9 && prevX > 2 && prevX < 8)
	{
		chain = manager_.GetChainAt(Point(prevX, 8));

		return (chain) && (chain->GetColor() == move.color) &&
			   (0 == (chainState_[chain->GetIndex()] & (WEAK | URGENT)) ) &&
			   (CheckFirstLineYoseHane(move, Point(prevX + 1, 9), Point(prevX - 1, 9), Point(prevX + 1, 8), Point(prevX - 1, 8)));
	}

	return false;
}

bool CapriccioAI::CheckFirstLineYoseHane(Move& move, Point pt1, Point pt2, Point pt3, Point pt4)
{
	if (IsStrictlySafe(Move(BLACK_STONECOLOR, pt1)) && IsStrictlySafe(Move(WHITE_STONECOLOR, pt1)))
	{
		move.point.x = pt1.x;
		move.point.y = pt1.y;
		return true;
	}
	else if (IsStrictlySafe(Move(BLACK_STONECOLOR, pt2)) && IsStrictlySafe(Move(WHITE_STONECOLOR, pt2)))
	{
		move.point.x = pt2.x;
		move.point.y = pt2.y;
		return true;
	}
	else if (IsStrictlySafe(Move(BLACK_STONECOLOR, pt3)) && IsStrictlySafe(Move(WHITE_STONECOLOR, pt3)))
	{
		move.point.x = pt3.x;
		move.point.y = pt3.y;
		return true;
	}
	else if (IsStrictlySafe(Move(BLACK_STONECOLOR, pt4)) && IsStrictlySafe(Move(WHITE_STONECOLOR, pt4)))
	{
		move.point.x = pt4.x;
		move.point.y = pt4.y;
		return true;
	}

	return false;
}

bool CapriccioAI::CheckSecondLineOsae(Move& move, Move osaeMove, Move oppMove1, Move oppMove2, Point chainPos)
{
	if (IsStrictlySafe(osaeMove))
	{
		StoneChain *haneChain = manager_.GetNewestStoneChain();
		BitBoard9  haneChainLibs;

		manager_.SetStone(osaeMove);
		move = osaeMove;

		StoneChain *chain = manager_.GetChainAt(chainPos);
		BitBoard9  chainLibs;
		manager_.GetLiberties(haneChain, &haneChainLibs);
		manager_.GetLiberties(chain, &chainLibs);

		bool ret = !CheckIfUrgent(chain) // if the chain above the hane move is still urgent after the response move, the osae is not valid
			       && (!IsStrictlySafe(oppMove1) || ( haneChainLibs.CountStones() < 3 && (chain->GetNumStones() > 1 || chainLibs.CountStones() > 2)));

		if (ret && IsStrictlySafe(oppMove2))
		{
			manager_.SetStone(oppMove2);
			ret = !(CheckIfWeak(chain) || CheckIfUrgent(chain));
			manager_.Undo();
		}

		manager_.Undo();

		return ret;
	}

	return false;
}

void CapriccioAI::AppendPossibleMoves(MoveList& moves, StoneColor color)
{
	Move move;
	move.color = color;

	for (move.point.x = 1; move.point.x <= 9; move.point.x++)
	{
		for (move.point.y = 1; move.point.y <= 9; move.point.y++)
		{
			if (manager_.IsLegalForBothColor(move.point))
			{
				moves.moves_[moves.count_].move_  = move;
				moves.moves_[moves.count_].value_ = 0;
				moves.count_++;
			}
		}
	}
}

// find capture move
bool CapriccioAI::FindAttackMove_0(Move &move, StoneChain *chain)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);

	return FindAttackMove_0_(move, libs, libs.CountStones());
}

// lower level function of FindAttackMove_0
bool CapriccioAI::FindAttackMove_0_(Move &move, BitBoard9 libs, int numLibs)
{
	if (numLibs == 1)
	{
		libs.GetTheOnlyPoint(&move.point);
		return manager_.IsLegal(move);
	}
	else
		return false;
}

// find ladder moves
bool CapriccioAI::FindAttackMove_1(MoveList& moves, StoneChain *chain)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);

	return FindAttackMove_1_(moves, chain, libs, libs.CountStones());
}

// lower level function of FindAttackMove_1
bool CapriccioAI::FindAttackMove_1_(MoveList& moves, StoneChain *chain, BitBoard9 libs, int numLibs)
{
	if (numLibs == 2)
	{
		BoardStonePointEnumerator libEnumerator;
		libEnumerator.Initialize(libs);
		Point* libPt;

		// temp move variables
		Move atkMove, defMove;
		defMove.color = chain->GetColor();
		atkMove.color = chain->GetColor() ^ 0x1;
		StoneChain* targetChain;

		bool ret = false;
		int  moveCount = moves.count_;

		while (libPt = libEnumerator.Next())
		{
			atkMove.point = *libPt;

			if (manager_.IsLegal(atkMove))
			{
				manager_.SetStone(atkMove);

				if (FindDefenseMove_1(defMove, chain))
				{
					manager_.SetStone(defMove);
					targetChain = manager_.GetNewestStoneChain();

					if (!targetChain->GetStones().HasStonesInCommon(chain->GetStones()))
					{
						targetChain = chain;
					}

					manager_.GetLiberties(targetChain, &libs);

					if (libs.CountStones() <= 1 || FindAttackMove_1_(moves, targetChain, libs, libs.CountStones()))
					{
						moves.moves_[moveCount].move_  = atkMove;
						moves.moves_[moveCount].value_ = chain->GetNumStones() << 2;
						moveCount++;

						ret = true;
					}

					manager_.Undo();
				}
				else
				{
					moves.moves_[moveCount].move_  = atkMove;
					moves.moves_[moveCount].value_ = chain->GetNumStones() << 2;
					moveCount++;
					ret = true;
				}

				manager_.Undo();
				moves.count_ = moveCount;
			}
		}

		return ret;

	}

	return false;
}

// find moves that enables (depth - 1) attacks
bool CapriccioAI::FindAttackMove(MoveList& moves, StoneChain* chain, int depth)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	int numLibs = libs.CountStones();

	Move move;
	move.color = chain->GetColor() ^ 0x1;

	if (depth <= 1 && numLibs <= 2)
	{
		if (FindAttackMove_0_(move, libs, numLibs))
		{
			moves.moves_[moves.count_].move_ = move;
			moves.moves_[moves.count_].value_ = chain->GetNumStones() << 1;
			moves.count_++;
			return true;
		}
		else if (FindAttackMove_1_(moves, chain, libs, numLibs))
		{
			return true;
		}
		else 
			return false;
	}
	else if (numLibs > depth)
	{
		return false;
	}
	// find shallow moves before deep moves
	else if (FindAttackMove(moves, chain, depth - 1))
	{
		return true;
	}
	else if (numLibs == depth)
	{
		BoardStonePointEnumerator libEnumerator;
		libEnumerator.Initialize(libs);
		Point* libPt;
		StoneChain* tempChain;
		Move oppMove, tempMove;
		tempMove.color = move.color;
		oppMove .color = move.color ^ 0x1;

		bool continueLoop;
		NeighborStoneChainEnumerator neighborEnumerator;
		StoneChain *neighbor;
		BitBoard9 neighborLibs;

		BitBoard9 newLibs;
		bool ret = false;
		int  moveCount = moves.count_;

		while (libPt = libEnumerator.Next())
		{
			move.point = *libPt;

			if (manager_.IsLegal(move))
			{
				manager_.SetStone(move);
				tempChain = manager_.GetNewestStoneChain();

				manager_.GetLiberties(tempChain, &newLibs);

				if (newLibs.CountStones() <= 2)
				{
					// if tempChain can be atari-ed, the attack is not valid
					continueLoop = true;

					neighborEnumerator = manager_.GetNeighborEnumeratorOf(tempChain);

					while (neighbor = neighborEnumerator.Next())
					{
						manager_.GetLiberties(neighbor, &neighborLibs);
						if (neighborLibs.CountStones() <= 2)
						{
							continueLoop = false;
							break;
						}
					}

					if (continueLoop)
					{
						manager_.Undo();
						continue;
					}
				}

				if ( !FindAttackMove(moves, tempChain, 1) &&
					FindAttackMove(moves, chain, depth - 1) )
				{
					moves.moves_[moveCount].move_  = move;
					moves.moves_[moveCount].value_ = chain->GetNumStones();
					moveCount++;

					ret = true;
				}
				moves.count_ = moveCount;
				manager_.Undo();
			}
		}

		return ret;
	}
	else // i.e. if (numLibs < depth)
	{
		// 取りあえずはfalseを返す
	}

	return false;
}

// find capture avoidance moves
bool CapriccioAI::FindDefenseMove_1(Move& move, StoneChain *chain)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);

	if (libs.CountStones() == 1)
	{
		// counter attack
		NeighborStoneChainEnumerator neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);;
		StoneChain* neighbor;

		while (neighbor = neighborEnumerator.Next())
		{
			if ( FindAttackMove_0(move, neighbor) )
			{
				return true;
			}
		}

		// simple escape
		libs.GetTheOnlyPoint(&move.point);
		return manager_.IsSafe(move);
	}
	else
		return false;

}

// find moves that disables (depth - 1) attacks
bool CapriccioAI::FindDefenseMove(MoveList &moves, StoneChain *chain, int depth)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	int numLibs = libs.CountStones();

	Move move;
	move.color = chain->GetColor();

	if (depth <= 1)
	{
		if (numLibs == 1)
		{
			if (FindDefenseMove_1(move, chain))
			{
				moves.moves_[moves.count_].move_  = move;
				moves.moves_[moves.count_].value_ = chain->GetNumStones() << 1;
				moves.count_++;

				return true;
			}
			return false;
		}
		else 
		{
			return true; // no need to defend against capture
		}
	}
	else if (numLibs > depth)
	{
		// too many libs to be a target of an attack w/ depth so shallow
		return true;
	}
	else
	{
		StoneChain* newChain;
		int moveCount = moves.count_;
		bool ret = false;

		if (FindAttackMove(moves, chain, depth - 1))
		{
			for (move.point.x = 1; move.point.x <= 9; move.point.x++)
			{
				for (move.point.y = 1; move.point.y <= 9; move.point.y++)
				{
					if (manager_.IsLegal(move))
					{
						manager_.SetStone(move);

						newChain = manager_.GetNewestStoneChain();
						if (!newChain->GetStones().HasStonesInCommon(chain->GetStones()))
						{
							// if newChain does not have any stones in common with the chain,
							// then the newChain is not the updated chain and thus it is chain that must be
							// searched for attack moves
							newChain = chain;
						}

						if (!FindAttackMove(moves, newChain, depth - 1))
						{
							// if all attack moves were cleared by this move, then this move was a valid defense move
							moves.moves_[moveCount].move_  = move;
							moves.moves_[moveCount].value_ = chain->GetNumStones();
							moveCount++;

							ret = true;
						}

						moves.count_ = moveCount;
						manager_.Undo();
					}
				}
			}

			moves.count_ = moveCount;

			if (ret)
				return ret;
		}

		// if no move was found, check simpler moves
		return FindDefenseMove(moves, chain, depth - 1);
	}
}

void CapriccioAI::FindSafeMakingMoves(MoveList &moves, StoneColor color)
{
	StoneChainEnumerator enumerator = manager_.GetEnumerator();
	StoneChain *chain, *newChain;
	NeighborStoneChainEnumerator neighborEnumerator;
	StoneChain* neighbor;

	Move move;
	bool continueLoop;
	int  listCount = searchList_.count_;

	BitBoard9 libs;
	int moveValue;

	while (chain = enumerator.Next())
	{
		if (chain->GetNumStones() > 3 && 0 == (chainState_[chain->GetIndex()] & SAFE) )
		{
			continueLoop = false;
			neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);

			while (neighbor = neighborEnumerator.Next())
			{
				if (0 != (chainState_[neighbor->GetIndex()] & (WEAK | URGENT)) )
				{
					// no need to make this chain safe
					continueLoop = true;
					break;
				}
			}

			if (continueLoop)
				continue;

			move.color = chain->GetColor();
			manager_.GetLiberties(chain, &libs);
			moveValue = 1 + chain->GetNumStones() - (libs.CountStones() >> 1); // more libs, the safer it is and thus less urgent

			if (moveValue > 0)
			{
				for (move.point.x = 1; move.point.x <= 9; move.point.x++)
				{
					for (move.point.y = 1; move.point.y <= 9; move.point.y++)
					{
						if (manager_.IsLegalForBothColor(move.point))
						{
							manager_.SetStone(move);

							newChain = manager_.GetNewestStoneChain();
							if (!newChain->GetStones().HasStonesInCommon(chain->GetStones()))
							{
								newChain = chain;
							}
							
							if (CheckIfSafe(newChain) && IsSafeWithOneLessLib(newChain))
							{
								searchList_.moves_[listCount].move_       = move;
								searchList_.moves_[listCount].move_.color = color;
								searchList_.moves_[listCount].value_      = moveValue;
								listCount = ++searchList_.count_ ;

								manager_.Undo();

								if (chain->GetColor() != color && !IsStrictlySafe(searchList_.moves_[listCount - 1].move_ ))
									listCount = --searchList_.count_;
							}
							else
								manager_.Undo();
						}
					}
				}
			}
		}
	}
}

void CapriccioAI::FindSafeBreakingMoves(MoveList &moves, StoneColor color)
{
	StoneChainEnumerator enumerator = manager_.GetEnumerator();
	StoneChain *chain;
	NeighborStoneChainEnumerator neighborEnumerator;
	StoneChain* neighbor;
	BoardStonePointEnumerator libEnumerator;
	Point* libPt;

	Move move, oppMove;
	bool continueLoop;
	int  listCount = searchList_.count_;

	BitBoard9 libs;
	int x, y;

	while (chain = enumerator.Next())
	{
		if (chain->GetColor() != color && (chainState_[chain->GetIndex()] & SAFE) && safeLibs_[chain->GetIndex()].CountStones() <= 2)
		{
			neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);

			continueLoop = false;
			while (neighbor = neighborEnumerator.Next())
			{
				if (chainState_[neighbor->GetIndex()] & (WEAK | URGENT))
				{
					// do something with the neighbors before trying out something new
					continueLoop = true;
				}
			}

			if (continueLoop )
				continue;

			move.color = chain->GetColor() ^ 0x01;
			oppMove.color = chain->GetColor() ^ 0x01;
			libEnumerator.Initialize(safeLibs_[chain->GetIndex()]);

			while (libPt = libEnumerator.Next())
			{
				x = libPt->x;
				y = libPt->y;

				move.point = *libPt;

				oppMove.point.x = x+1;
				oppMove.point.y = y;
				if (manager_.IsLegal(oppMove))
				{
					manager_.SetStone(oppMove);

					if (!CheckIfSafe(chain) && !CheckSafeConnection(chain))
					{
						searchList_.moves_[listCount].move_  = oppMove;
						searchList_.moves_[listCount].move_.color = color;
						searchList_.moves_[listCount].value_ = chain->GetNumStones() >> 1;
						searchList_.count_ = ++listCount;
					}

					manager_.Undo();
				}

				oppMove.point.x = x-1;
				oppMove.point.y = y;
				if (manager_.IsLegal(oppMove))
				{
					manager_.SetStone(oppMove);

					if (!CheckIfSafe(chain) && !CheckSafeConnection(chain))
					{
						searchList_.moves_[listCount].move_  = oppMove;
						searchList_.moves_[listCount].move_.color = color;
						searchList_.moves_[listCount].value_ = chain->GetNumStones() >> 1;
						searchList_.count_ = ++listCount;
					}

					manager_.Undo();
				}

				oppMove.point.x = x;
				oppMove.point.y = y+1;
				if (manager_.IsLegal(oppMove))
				{
					manager_.SetStone(oppMove);

					if (!CheckIfSafe(chain) && !CheckSafeConnection(chain))
					{
						searchList_.moves_[listCount].move_  = oppMove;
						searchList_.moves_[listCount].move_.color = color;
						searchList_.moves_[listCount].value_ = chain->GetNumStones() >> 1;
						searchList_.count_ = ++listCount;
					}

					manager_.Undo();
				}

				oppMove.point.x = x;
				oppMove.point.y = y-1;
				if (manager_.IsLegal(oppMove))
				{
					manager_.SetStone(oppMove);

					if (!CheckIfSafe(chain) && !CheckSafeConnection(chain))
					{
						searchList_.moves_[listCount].move_  = oppMove;
						searchList_.moves_[listCount].move_.color = color;
						searchList_.moves_[listCount].value_ = chain->GetNumStones() >> 1;
						searchList_.count_ = ++listCount;
					}

					manager_.Undo();
				}
			}
		}
	}
}

void CapriccioAI::FindYoseMoves(MoveList &moves, StoneColor color)
{
	Move move;
	move.color = color;
	StoneColor oppColor = color ^ 0x01;
	bool continueLoop;
	NeighborStoneChainEnumerator neighborEnumerator;
	StoneChain* neighbor;

	BitBoard9 libs;

	// first line yose
	StoneChain* chainAbove;
	Point ptAbove;
	ptAbove.x = 2;
	move.point.x = 1;

	for (move.point.y = 2; move.point.y <= 8; move.point.y++)
	{
		if (IsStrictlySafe(move))
		{
			ptAbove.y = move.point.y;
			chainAbove = manager_.GetChainAt(ptAbove);

			if (chainAbove && (0 == (chainState_[chainAbove->GetIndex()] & (WEAK | URGENT))) &&
				chainAbove->GetColor() != color)
			{
				manager_.SetStone(move);

				manager_.GetLiberties(manager_.GetNewestStoneChain(), &libs);

				moves.moves_[moves.count_].move_  = move;
				moves.moves_[moves.count_].value_ = libs.CountStones() << (int)(manager_.HasAtariChainAround(move));
				moves.count_++;

				manager_.Undo();
			}
		}
	}

	move.point.x = 9;
	ptAbove.x = 8;
	for (move.point.y = 2; move.point.y <= 8; move.point.y++)
	{
		if (IsStrictlySafe(move))
		{
			ptAbove.y = move.point.y;
			chainAbove = manager_.GetChainAt(ptAbove);

			if (chainAbove && (0 == (chainState_[chainAbove->GetIndex()] & (WEAK | URGENT))) &&
				chainAbove->GetColor() != color)
			{
				manager_.SetStone(move);

				manager_.GetLiberties(manager_.GetNewestStoneChain(), &libs);

				moves.moves_[moves.count_].move_  = move;
				moves.moves_[moves.count_].value_ = libs.CountStones() << (int)(manager_.HasAtariChainAround(move));
				moves.count_++;

				manager_.Undo();
			}
		}
	}

	move.point.y = 1;
	ptAbove.y = 2;
	for (move.point.x = 2; move.point.x <= 8; move.point.x++)
	{
		if (IsStrictlySafe(move))
		{
			ptAbove.x = move.point.x;
			chainAbove = manager_.GetChainAt(ptAbove);

			if (chainAbove && (0 == (chainState_[chainAbove->GetIndex()] & (WEAK | URGENT))) &&
				chainAbove->GetColor() != color)
			{
				manager_.SetStone(move);

				manager_.GetLiberties(manager_.GetNewestStoneChain(), &libs);

				moves.moves_[moves.count_].move_  = move;
				moves.moves_[moves.count_].value_ = libs.CountStones() << (int)(manager_.HasAtariChainAround(move));
				moves.count_++;

				manager_.Undo();
			}
		}
	}

	move.point.y = 9;
	ptAbove.y = 8;
	for (move.point.x = 2; move.point.x <= 8; move.point.x++)
	{
		if (IsStrictlySafe(move))
		{
			ptAbove.x = move.point.x;
			chainAbove = manager_.GetChainAt(ptAbove);

			if (chainAbove && (0 == (chainState_[chainAbove->GetIndex()] & (WEAK | URGENT))) &&
				chainAbove->GetColor() != color)
			{
				manager_.SetStone(move);

				manager_.GetLiberties(manager_.GetNewestStoneChain(), &libs);

				moves.moves_[moves.count_].move_  = move;
				moves.moves_[moves.count_].value_ = libs.CountStones() << (int)(manager_.HasAtariChainAround(move));
				moves.count_++;

				manager_.Undo();
			}
		}
	}

	for (move.point.x = 2; move.point.x <= 8; move.point.x++)
	{
		for (move.point.y = 2; move.point.y <= 8; move.point.y++)
		{
			if (IsStrictlySafe(move))
			{
				if (manager_.HasBothColorStoneAround(move.point))
				{
					neighborEnumerator = manager_.GetNeighborEnumeratorAround(move.point);
					continueLoop = false;

					while(neighbor = neighborEnumerator.Next())
					{
						if (chainState_[neighbor->GetIndex()] & (WEAK | URGENT))
						{
							continueLoop = true;
							break;
						}
					}
					if (continueLoop)
						continue;

					manager_.SetStone(move);

					manager_.GetLiberties(manager_.GetNewestStoneChain(), &libs);

					moves.moves_[moves.count_].move_  = move;
					moves.moves_[moves.count_].value_ = libs.CountStones() << (int)(manager_.HasAtariChainAround(move));
					moves.count_++;

					manager_.Undo();
				}
				else
				{
					unsigned char x = move.point.x;
					unsigned char y = move.point.y;

					if (manager_.HasStone(Point(x + 1, y), oppColor) || manager_.HasStone(Point(x - 1, y), oppColor))
					{
						manager_.SetStone(move);

						if (!IsStrictlySafe(Move(oppColor, x, y + 1)) || !IsStrictlySafe(Move(oppColor, x, y - 1)))
						{
							manager_.GetLiberties(manager_.GetNewestStoneChain(), &libs);

							moves.moves_[moves.count_].move_  = move;
							moves.moves_[moves.count_].value_ = libs.CountStones() << (int)(manager_.HasAtariChainAround(move));
							moves.count_++;
						}

						manager_.Undo();
					}

					if (manager_.HasStone(Point(x, y + 1), oppColor) || manager_.HasStone(Point(x, y - 1), oppColor))
					{
						manager_.SetStone(move);

						if (!IsStrictlySafe(Move(oppColor, x + 1, y)) || !IsStrictlySafe(Move(oppColor, x - 1, y)))
						{
							manager_.GetLiberties(manager_.GetNewestStoneChain(), &libs);

							moves.moves_[moves.count_].move_  = move;
							moves.moves_[moves.count_].value_ = libs.CountStones() << (int)(manager_.HasAtariChainAround(move));
							moves.count_++;
						}

						manager_.Undo();
					}
				}
			}
		}
	}
}

int CapriccioAI::GetRelativity(Point pt, StoneChain *chain)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	int numLibs = libs.CountStones();

	int distance = 10;

	// distance 1 is bit special... 
	BitBoard9 distMask = distanceMask_[pt.x][pt.y][0];
	distMask &= chain->GetStones();
	int stoneCount = distMask.CountStones();

	if (stoneCount == 1)
	{
		// set relativity, reflecting the size of stone chain
		return 1 + (chain->GetNumStones() >> 1);
	}
	else if (stoneCount > 1)
	{
		// possible guzumi move, try to avoid it
		return -5;
	}

	for (int d = 1; d < MAX_RELATIVE_DISTANCE; d++)
	{
		if (distanceMask_[pt.x][pt.y][d].HasStonesInCommon(chain->GetStones()))
		{
			distance = d + 1;
			distMask  = distanceMask_[pt.x][pt.y][d];
			distMask &= chain->GetStones();
			distMask.Update();
			break;
		}
	}

	if (distance == 10 || numLibs <= 2)
	{
		// no stones found... (too far from the point) OR
		// possible chance for atari makes the relativity (or actually the importance of the relativity) low
		return 0;
	}

	int relativity = 0;

	BoardStonePointEnumerator ptEnumerator ;
	ptEnumerator.Initialize(distMask);
	Point* stonePt;

	int minX, maxX, minY, maxY, distX, distY;

	while (stonePt = ptEnumerator.Next())
	{
		relativity += 16 - (distance << 1);

		if (stonePt->x < pt.x)
		{
			minX = stonePt->x - 1;
			maxX = pt.x + 1;
		}
		else
		{
			minX = pt.x - 1;
			maxX = stonePt->x + 1;
		}
		distX = maxX - minX;

		if (stonePt->y < pt.y)
		{
			minY = stonePt->y - 1;
			maxY = pt.y + 1;
		}
		else
		{
			minY = pt.y - 1;
			maxY = stonePt->y + 1;
		}
		distY = maxY - minY;

		if (distX > distY)
		{
			relativity -= distY;
			if (distX == 4)
				relativity -= 1;
			else if (distX == 5)
				relativity -= 3;
			else if (distX == 6)
				relativity -= 5;
		}
		else
		{
			relativity -= distX;
			if (distY == 4)
				relativity -= 1;
			else if (distY == 5)
				relativity -= 3;
			else if (distY == 6)
				relativity -= 5;
		}

		StoneColor color = chain->GetColor() ^ 0x1;



		if (distX == distY && distance == 2)
		{
			// this shape makes a soundly shaped cut
			for (int x = minX+1; x < maxX; x++)
			{
				for (int y = minY+1; y < maxY; y++)
				{
					if (manager_.HasStone(Point(x, y), color))
					{
						relativity += distance;
					}
				}
			}
		}
		else
		{
			// stones are interrupting the relativity
			for (int x = minX; x <= maxX; x++)
			{
				for (int y = minY; y <= maxY; y++)
				{
					if (manager_.HasStone(Point(x, y), color))
					{
						if (x - minX == maxX - x || y - minY == maxY - y)
							relativity -= (distance << 1) + 1;
						else
							relativity -= distance;
					}
				}
			}
		}
	}

	// more calculations?

	if ( relativity < 0)
		return 0;
	else
		return relativity;

}

void CapriccioAI::UpdateChainState()
{
	StoneChainEnumerator enumerator = manager_.GetEnumerator();
	StoneChain* chain;
	NeighborStoneChainEnumerator neighbotEnumerator;
	StoneChain* neighbor;

	BitBoard9 libs;
	Move move;


	while (chain = enumerator.Next())
	{
		chainState_[chain->GetIndex()] = NONE;
		if (CheckIfSafe(chain))
		{
			chainState_[chain->GetIndex()] = SAFE;
		}
		else if (CheckIfWeak(chain))
		{
			chainState_[chain->GetIndex()] = WEAK;
		}
		else if (CheckIfUrgent(chain))
		{
			chainState_[chain->GetIndex()] = URGENT;
		}
		else if (CheckSafeConnection(chain))
		{
			chainState_[chain->GetIndex()] = SAFE;
		}
	}

	enumerator = manager_.GetEnumerator();

	// go through the chains once again to check 
	// if the neighbors of weak chains are weak or not
	while (chain = enumerator.Next())
	{
		if (chainState_[chain->GetIndex()] & (WEAK | URGENT))
		{
			neighbotEnumerator = manager_.GetNeighborEnumeratorOf(chain);

			while (neighbor = neighbotEnumerator.Next())
			{
				if (chainState_[neighbor->GetIndex()] & (WEAK | URGENT)) 
				{
					manager_.GetLiberties(neighbor, &libs);

					// check for snapback
					if (libs.CountStones() == 1)
					{
						libs.GetTheOnlyPoint(&move.point);
						move.color = chain->GetColor();
						if (!manager_.IsSafe(move))
						{
							continue;
						}
					}

					chainState_[chain->GetIndex()] = URGENT;
				}
			}
		}
	}
}

bool CapriccioAI::CheckIfWeak(StoneChain* chain)
{
	// if the chain cannot escape, it is weak

	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	int numLibs = libs.CountStones();

	Move move;
	move.color = chain->GetColor();

	if (numLibs == 1)
	{
		if (FindDefenseMove_1(move, chain))
		{
			manager_.SetStone(move);

			StoneChain* newChain = manager_.GetNewestStoneChain();
			if (!newChain->GetStones().HasStonesInCommon(chain->GetStones()))
			{
				newChain = chain;
			}

			int listCount = searchList_.count_;
			bool ret = FindAttackMove(searchList_, newChain, 1);

			manager_.Undo();
			searchList_.count_ = listCount;
			return ret;
		}
		else
		{
			return true;
		}
	}
	else if (numLibs == 2)
	{
		Move libMove1, libMove2;
		int listCount;

		BoardStonePointEnumerator libEnumerator;
		libEnumerator.Initialize(libs);

		libMove1.point = *(libEnumerator.Next());
		libMove1.color = move.color;
		libMove2.point = *(libEnumerator.Next());
		libMove2.color = move.color;

		// try playing on the first lib
		{
			manager_.SetStone(libMove1);
			listCount = searchList_.count_;

			if (!FindAttackMove(searchList_, manager_.GetNewestStoneChain(), 1))
			{
				// if playing on this lib disables an immediate capture, this chain isn't weak
				manager_.Undo();
				return false;
			}

			manager_.Undo();
			searchList_.count_ = listCount;
		}

		// try playing on the second lib
		{
			manager_.SetStone(libMove2);
			listCount = searchList_.count_;

			if (!FindAttackMove(searchList_, manager_.GetNewestStoneChain(), 1))
			{
				// if playing on this lib disables an immediate capture, this chain isn't weak
				manager_.Undo();
				return false;
			}

			manager_.Undo();
			searchList_.count_ = listCount;
		}

		return true;
	}

	return false;
}

bool CapriccioAI::CheckIfUrgent(StoneChain *chain)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	int numLibs = libs.CountStones();

	if (numLibs == 1)
		return true;
	else if (numLibs > 3)
		return false;

	NeighborStoneChainEnumerator neighborEnumerator;
	StoneChain *neighbor;
	BitBoard9  neighborLibs;
	int listCount = searchList_.count_;

	BoardStonePointEnumerator libEnumerator;
	Point* libPoint;
	Move move;
	move.color = chain->GetColor() ^ 0x01;

	neighborEnumerator = manager_.GetNeighborEnumeratorOf(chain);

	while (neighbor = neighborEnumerator.Next())
	{
		manager_.GetLiberties(neighbor, &neighborLibs);
		if (CheckIfWeak(neighbor) )
		{
			searchList_.count_ = listCount;
			return false;
		}
	}

	searchList_.count_ = listCount;

	libEnumerator.Initialize(libs);

	while (libPoint = libEnumerator.Next())
	{
		move.point = *libPoint;

		if (manager_.IsLegal(move))
		{
			manager_.SetStone(move);

			if (CheckIfWeak(chain))
			{
				manager_.Undo();
				return true;
			}

			manager_.Undo();
		}
	}

	return false;
}

bool CapriccioAI::CheckIfSafe(StoneChain *chain)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	int numLibs = libs.CountStones();

	BoardStonePointEnumerator libEnumerator;

	Move move;
	move.color = chain->GetColor() ^ 0x1;

	safeLibs_[chain->GetIndex()].Clear();

	if (numLibs == 1)
	{
		return false;	// of course, atari-ed chains are not safe
	}

	// special case for a single stone on the edge
	if (chain->GetNumStones() == 1 && chain->GetStones().IsEdgeLevelRank())
	{
		if (numLibs == 2)
		{
			// safe if all possible atari can be cleared in the next move
			Move libMove1, libMove2;
			libEnumerator.Initialize(libs);

			libMove1.point = *(libEnumerator.Next());
			libMove1.color = chain->GetColor() ^ 0x1;
			libMove2.point = *(libEnumerator.Next());
			libMove2.color = libMove1.color;

			return !manager_.IsSafe(libMove1) && !manager_.IsSafe(libMove2);
		}
		else
		{
			// safe if all liberty-decreasing moves can be captured or laddered
			libEnumerator.Initialize(libs);
			Point* libPoint;

			while (libPoint = libEnumerator.Next())
			{
				move.point = *libPoint;

				if (IsStrictlySafe(move))
					return false;
			}

			return true;
		}	
	}

	libEnumerator.Initialize(libs);
	Point* libPoint;
	int listCount = searchList_.count_;
	int numSafeLibs = 0;

	while (libPoint = libEnumerator.Next())
	{
		move.point = *libPoint;

		if (!IsStrictlySafe(move))
		{
			searchList_.moves_[searchList_.count_++].move_ = move;

			safeLibs_[chain->GetIndex()].SetStone(move.point);

			if (++numSafeLibs > 1)
			{
				for (int i = searchList_.count_ - 1; i >= listCount; --i)
				{
					if (IsSafeLibertyPair(move.point, searchList_.moves_[i].move_.point, chain->GetColor()))
					{
						searchList_.count_ = listCount;
						safeLibs_[chain->GetIndex()].Update();

						return true;
					}
				}
			}
		}
	}

	searchList_.count_ = listCount;
	safeLibs_[chain->GetIndex()].Update();

	return false;
}

// note: make sure to call this function after calling CheckIfSafe on the same chain
bool CapriccioAI::CheckSafeConnection(StoneChain* chain)
{
	BoardStonePointEnumerator enumerator;
	Point* safeLibPt;
	Move move;
	StoneChain* newChain;
	BitBoard9 libs;
	int numStones, newNumStones;;

	int searchListCount = searchList_.count_;
	if (FindAttackMove(searchList_, chain, 1))
	{
		searchList_.count_ = searchListCount;
		return false;
	}
	searchList_.count_ = searchListCount;

	move.color = chain->GetColor();
	enumerator.Initialize(safeLibs_[chain->GetIndex()]);
	numStones = chain->GetNumStones() + 1; // if the chain size only grew by 1, it didn't connect to anything

	while (safeLibPt = enumerator.Next())
	{
		move.point = *safeLibPt;

		manager_.SetStone(move);
		newChain = manager_.GetNewestStoneChain();
		newNumStones = newChain->GetNumStones();

		manager_.GetLiberties(chain, &libs);
		if (numStones < newNumStones && (newNumStones> 5 && libs.CountStones() > 4 || CheckIfSafe(newChain)))
		{
			manager_.Undo();
			return true;
		}

		manager_.Undo();
	}

	return false;
}

bool CapriccioAI::IsSafeLibertyPair(Point pt1, Point pt2, StoneColor color)
{
	int minX, maxX, minY, maxY;

	if (pt1.x < pt2.x)
	{
		minX = pt1.x;
		maxX = pt2.x;
	}
	else
	{
		minX = pt2.x;
		maxX = pt1.x;
	}

	if (pt1.y < pt2.y)
	{
		minY = pt1.y;
		maxY = pt2.y;
	}
	else
	{
		minY = pt2.y;
		maxY = pt1.y;
	}

	int dist = maxX + maxY - minX - minY;

	if (dist > 3)
	{
		return true;
	}
	else if (dist == 3)
	{
		int distX = maxX - minX;
		int distY = maxY - minY;

		// find the "L 4-space shape"
		// if that shape exists in the territory, there is a possible chance of nakade

		if (distX == 1)			// i.e distY == 2
		{
			if ( (!manager_.HasStone(Point(maxX, minY    ), color) &&
				  !manager_.HasStone(Point(maxX, minY + 1), color) &&
				  !manager_.HasStone(Point(maxX, maxY    ), color) )  ||
				 (!manager_.HasStone(Point(minX, minY    ), color) &&
				  !manager_.HasStone(Point(minX, minY + 1), color) &&
				  !manager_.HasStone(Point(minX, maxY    ), color) )  )
			{
				return false;
			}
		}
		else if (distY == 1)	// i.e distX == 2
		{
			if ( (!manager_.HasStone(Point(minX    , maxY), color) &&
				  !manager_.HasStone(Point(maxX + 1, maxY), color) &&
				  !manager_.HasStone(Point(maxY    , maxY), color) )  ||
				 (!manager_.HasStone(Point(minX    , minY), color) &&
				  !manager_.HasStone(Point(minX + 1, minY), color) &&
				  !manager_.HasStone(Point(maxX    , minY), color) )  )
			{
				return false;
			}
		}
		else if (distX == 0)
		{
			if (!manager_.HasStone(Point(minX, minY + 1), color) &&
				!manager_.HasStone(Point(minX, maxY - 1), color) )
			{
				return false;
			}
		}
		else if (distY == 0)
		{
			if (!manager_.HasStone(Point(minX + 1, minY), color) &&
				!manager_.HasStone(Point(maxX - 1, minY), color) )
			{
				return false;
			}
		}

		return true;
	}
	else if (dist == 2)
	{
		// find 3-pt line space separated by a stone
		// or a kosumi shape separated by a kosumi shaped stones

		if ( (minX == maxX && manager_.HasStone(Point(minX, minY + 1), color)) ||
			 (minY == maxY && manager_.HasStone(Point(minX + 1, minY), color)) ||
			 (manager_.IsEmpty(Point(minX, maxY)) &&
			  manager_.HasStone(Point(minX, minY), color) &&
			  manager_.HasStone(Point(maxX, maxY), color) )		||
			 (manager_.HasStone(Point(minX, maxY), color) &&
			  manager_.HasStone(Point(maxX, minY), color) )     )
		{
			return true;
		}
		else
			return false;
	}
	else
	{
		return false;
	}
}

bool CapriccioAI::IsSafeWithOneLessLib(StoneChain* chain)
{
	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	BoardStonePointEnumerator libEnumerator;
	libEnumerator.Initialize(libs);
	Point* libPoint;
	Move move;
	move.color = chain->GetColor() ^ 0x01;
	StoneChain *newChain;

	while (libPoint = libEnumerator.Next())
	{
		move.point = *libPoint;

		if (manager_.IsSafe(move))
		{
			manager_.SetStone(move);

			newChain = manager_.GetNewestStoneChain();
			if (!newChain->GetStones().HasStonesInCommon(chain->GetStones()))
			{
				newChain = chain;
			}

			if (!CheckIfSafe(newChain))
			{
				manager_.Undo();
				return false;
			}

			manager_.Undo();
		}
	}

	return true;
}

bool CapriccioAI::IsLooseLadder(StoneChain *chain)
{
	// assume that the chain sent in as an argument is in urgent state

	BitBoard9 libs;
	manager_.GetLiberties(chain, &libs);
	BoardStonePointEnumerator libEnumerator;
	libEnumerator.Initialize(libs);
	Point* libPoint;
	Move move;
	move.color = chain->GetColor();
	StoneChain *newChain;
	BitBoard9 newLibs;
	int numNewLibs;
	BoardStonePointEnumerator newLibEnumerator;
	Move oppMove;
	oppMove.color = move.color ^ 0x01;

	bool returnFalse = false;

	while (libPoint = libEnumerator.Next())
	{
		move.point = *libPoint;

		if (manager_.IsSafe(move))
		{
			manager_.SetStone(move);
			newChain = manager_.GetNewestStoneChain();

			manager_.GetLiberties(newChain, &newLibs);
			numNewLibs = newLibs.CountStones();

			if (numNewLibs > 3 || CheckIfSafe(newChain))
			{
				// if there are 4 or more libs, the chain is of course not in loose ladder
				manager_.Undo();
				return false;
			}

			// find at lease one move that makes valid loose ladder
			newLibEnumerator.Initialize(newLibs);
			returnFalse = true;
			while (returnFalse && (libPoint = newLibEnumerator.Next()))
			{
				oppMove.point = *libPoint;

				if (IsStrictlySafe(oppMove))
				{
					manager_.SetStone(oppMove);
					//if ( CheckIfUrgent(newChain))
					//	returnFalse = !IsLooseLadder(newChain);
					returnFalse = !CheckIfUrgent(newChain);
					manager_.Undo();
				}
			}

			manager_.Undo();
			
			if (returnFalse)
				return false;
		}
	}

	return true;
}

bool CapriccioAI::IsStrictlySafe(Move move)
{
	// a move is strictly safe if the move cannot be instantly captured or laddered

	if (manager_.IsSafe(move))
	{
		manager_.SetStone(move);
		int listCount = searchList_.count_;

		bool ret = !FindAttackMove(searchList_, manager_.GetNewestStoneChain(), 1);

		searchList_.count_ = listCount;
		manager_.Undo();
		return ret;
	}

	return false;
}

#if defined (ENABLE_DEBUG_OUTPUT)

void CapriccioAI::ShowAllChains()
{
	searchList_.count_ = 0;
	UpdateChainState();

	char* comments[MAX_UNDOS];

	StoneChainEnumerator enumerator = manager_.GetEnumerator();
	StoneChain* chain;

	char text[MAX_UNDOS * 16];
	int textOffset = 0;

	while (chain = enumerator.Next())
	{

		switch(chainState_[chain->GetIndex()])
		{
		case NONE:
			sprintf(text + textOffset, "State = None");
			break;
		case SAFE:
			sprintf(text + textOffset, "State = Safe");
			break;
		case WEAK:
			sprintf(text + textOffset, "State = Weak");
			break;
		case URGENT:
			sprintf(text + textOffset, "State = Urgent");
			break;
		default:
			sprintf(text + textOffset, "State = %x", chainState_[chain->GetIndex()]);
		}

		comments[chain->GetIndex()] = text + textOffset;
		textOffset += 128;
	}

	manager_.ShowAllChains((char**)comments);
}

#endif